from __future__ import annotations

import base64
import datetime
import decimal
import ipaddress
import re
import urllib.parse
import uuid

from . import t, types

if t.TYPE_CHECKING:
    from .errors import ParseFragment


def toNative(val: t.Value, pf: ParseFragment) -> t.KDLishValue:
    if isinstance(val, types.Numberish):
        if val.tag == "i8":
            return i8(val, pf)
        if val.tag == "i16":
            return i16(val, pf)
        if val.tag == "i32":
            return i32(val, pf)
        if val.tag == "i64":
            return i64(val, pf)
        if val.tag == "u8":
            return u8(val, pf)
        if val.tag == "u16":
            return u16(val, pf)
        if val.tag == "u32":
            return u32(val, pf)
        if val.tag == "u64":
            return u64(val, pf)
        if val.tag in ("f32", "f64"):
            return int(val.value)
        if val.tag in ("decimal64", "decimal128"):
            return decim(val, pf)
    if isinstance(val, types.Stringish):
        if val.tag == "date-time":
            return dateTime(val, pf)
        if val.tag == "time":
            return time(val, pf)
        if val.tag == "date":
            return date(val, pf)
        if val.tag == "decimal":
            return decim(val, pf)
        if val.tag == "ipv4":
            return ipv4(val, pf)
        if val.tag == "ipv6":
            return ipv6(val, pf)
        if val.tag == "url":
            return url(val, pf)
        if val.tag == "uuid":
            return _uuid(val, pf)
        if val.tag == "regex":
            return regex(val, pf)
        if val.tag == "base64":
            return b64(val, pf)
    return val


def i8(val: t.Numberish, pf: ParseFragment) -> int:
    limit = 2**7
    if not (-limit <= val.value < limit):
        msg = f"{val.value} doesn't fit in an i8."
        raise pf.error(msg)
    return int(val.value)


def i16(val: t.Numberish, pf: ParseFragment) -> int:
    limit = 2**15
    if not (-limit <= val.value < limit):
        msg = f"{val.value} doesn't fit in an i16."
        raise pf.error(msg)
    return int(val.value)


def i32(val: t.Numberish, pf: ParseFragment) -> int:
    limit = 2**31
    if not (-limit <= val.value < limit):
        msg = f"{val.value} doesn't fit in an i32."
        raise pf.error(msg)
    return int(val.value)


def i64(val: t.Numberish, pf: ParseFragment) -> int:
    limit = 2**63
    if not (-limit <= val.value < limit):
        msg = f"{val.value} doesn't fit in an i64."
        raise pf.error(msg)
    return int(val.value)


def u8(val: t.Numberish, pf: ParseFragment) -> int:
    limit = 2**8
    if not (0 <= val.value < limit):
        msg = f"{val.value} doesn't fit in a u8."
        raise pf.error(msg)
    return int(val.value)


def u16(val: t.Numberish, pf: ParseFragment) -> int:
    limit = 2**16
    if not (0 <= val.value < limit):
        msg = f"{val.value} doesn't fit in a u16."
        raise pf.error(msg)
    return int(val.value)


def u32(val: t.Numberish, pf: ParseFragment) -> int:
    limit = 2**32
    if not (0 <= val.value < limit):
        msg = f"{val.value} doesn't fit in a u32."
        raise pf.error(msg)
    return int(val.value)


def u64(val: t.Numberish, pf: ParseFragment) -> int:
    limit = 2**64
    if not (0 <= val.value < limit):
        msg = f"{val.value} doesn't fit in a u64."
        raise pf.error(msg)
    return int(val.value)


def decim(val: t.Value, pf: ParseFragment) -> decimal.Decimal:
    if isinstance(val, types.Numberish):
        chars = pf.fragment.replace("_", "")
    else:
        chars = val.value
    try:
        return decimal.Decimal(chars)
    except decimal.InvalidOperation as exc:
        msg = f"Couldn't parse a decimal from {pf.fragment}."
        raise pf.error(msg) from exc


def dateTime(val: t.Stringish, pf: ParseFragment) -> datetime.datetime:
    try:
        return datetime.datetime.fromisoformat(val.value)
    except ValueError as exc:
        msg = f"Couldn't parse a date-time from {pf.fragment}."
        raise pf.error(msg) from exc


def time(val: t.Stringish, pf: ParseFragment) -> datetime.time:
    try:
        return datetime.time.fromisoformat(val.value)
    except ValueError as exc:
        msg = f"Couldn't parse a date-time from {pf.fragment}."
        raise pf.error(msg) from exc


def date(val: t.Stringish, pf: ParseFragment) -> datetime.date:
    try:
        return datetime.date.fromisoformat(val.value)
    except ValueError as exc:
        msg = f"Couldn't parse a date from {pf.fragment}."
        raise pf.error(msg) from exc


def ipv4(val: t.Stringish, pf: ParseFragment) -> ipaddress.IPv4Address:
    try:
        return ipaddress.IPv4Address(val.value)
    except ipaddress.AddressValueError as exc:
        msg = f"Couldn't parse an IPv4 address from {pf.fragment}."
        raise pf.error(msg) from exc


def ipv6(val: t.Stringish, pf: ParseFragment) -> ipaddress.IPv6Address:
    try:
        return ipaddress.IPv6Address(val.value)
    except ipaddress.AddressValueError as exc:
        msg = f"Couldn't parse an IPv6 address from {pf.fragment}."
        raise pf.error(msg) from exc


def url(val: t.Stringish, pf: ParseFragment) -> urllib.parse.ParseResult:
    try:
        return t.cast("urllib.parse.ParseResult", urllib.parse.urlparse(val.value))
    except ValueError as exc:
        msg = f"Couldn't parse a url from {pf.fragment}."
        raise pf.error(msg) from exc


def _uuid(val: t.Stringish, pf: ParseFragment) -> uuid.UUID:
    try:
        return uuid.UUID(val.value)
    except Exception as exc:
        msg = f"Couldn't parse a UUID from {pf.fragment}."
        raise pf.error(msg) from exc


def regex(val: t.Stringish, pf: ParseFragment) -> re.Pattern:
    try:
        return re.compile(val.value)
    except Exception as exc:
        msg = f"Couldn't parse a regex from {pf.fragment}."
        raise pf.error(msg) from exc


def b64(val: t.Stringish, pf: ParseFragment) -> bytes:
    try:
        return base64.b64decode(val.value.encode("utf-8"), validate=True)
    except Exception as exc:
        msg = "Couldn't parse base64."
        raise pf.error(msg) from exc



from __future__ import annotations

from typing import NamedTuple

from . import t


class Failure:
    pass


class Result(NamedTuple):
    value: t.Any
    end: int

    @property
    def valid(self) -> bool:
        return self.value is not Failure

    @staticmethod
    def fail(index: int) -> Result:
        return Result(Failure, index)



import typing

from . import parsing, printing
from .cli import cli
from .errors import ParseError, ParseFragment
from .parsefuncs import parse
from .parsing import ParseConfig, Parser
from .printing import PrintConfig
from .types import (
    Binary,
    Bool,
    Decimal,
    Document,
    ExactValue,
    Hex,
    Node,
    Null,
    Numberish,
    Octal,
    RawString,
    String,
    Stringish,
    Value,
)

if typing.TYPE_CHECKING:
    from .t import KDLAny, KDLishValue, KDLValue, NodeKey



# pylint: skip-file
# Module for holding types, for easy importing into the rest of the codebase
from __future__ import annotations

# The only three things that should be available during runtime.
from typing import TYPE_CHECKING, cast, overload

if TYPE_CHECKING:
    from typing import (
        AbstractSet,
        Any,
        AnyStr,
        Awaitable,
        Callable,
        DefaultDict,
        Deque,
        FrozenSet,
        Generator,
        Generic,
        Iterable,
        Iterator,
        Literal,
        Mapping,
        MutableMapping,
        MutableSequence,
        NamedTuple,
        NewType,
        Protocol,
        Sequence,
        TextIO,
        TypeAlias,
        TypedDict,
        TypeGuard,
        TypeVar,
    )

    from _typeshed import SupportsKeysAndGetItem
    from typing_extensions import (
        NotRequired,
        Required,
    )

    from .parsing import ParseConfig
    from .printing import PrintConfig
    from .stream import Stream
    from .types import (
        Binary,
        Bool,
        Decimal,
        Document,
        ExactValue,
        Hex,
        Node,
        Null,
        Numberish,
        Octal,
        RawString,
        String,
        Stringish,
        Value,
    )

    NodeKey = str | tuple[str | None, str | None]

    KDLAny: TypeAlias = Document | Node | KDLValue
    KDLValue: TypeAlias = (
        Value | Binary | Bool | Decimal | ExactValue | Hex | Null | Numberish | Octal | RawString | String | Stringish
    )

    import datetime
    import decimal
    import ipaddress
    import re
    import urllib
    import uuid

    KDLishValue: TypeAlias = (
        KDLValue
        | None
        | bool
        | String
        | int
        | float
        | decimal.Decimal
        | datetime.time
        | datetime.date
        | datetime.datetime
        | ipaddress.IPv4Address
        | ipaddress.IPv6Address
        | urllib.parse.ParseResult
        | uuid.UUID
        | re.Pattern
        | bytes
    )



import argparse
import sys

from . import parsefuncs, parsing, printing


def cli() -> None:
    cliParser = argparse.ArgumentParser(
        description="KDL parser/printer, letting you easily reformat KDL files into a canonical representation.",
    )
    cliParser.add_argument(
        "infile",
        nargs="?",
        type=argparse.FileType("r", encoding="utf-8"),
        default=sys.stdin,
    )
    cliParser.add_argument(
        "outfile",
        nargs="?",
        type=argparse.FileType("w", encoding="utf-8"),
        default=sys.stdout,
    )
    cliParser.add_argument(
        "--indent",
        dest="indent",
        type=int,
        action="store",
        default="-1",
        help="How many spaces for each level of indent. -1 indicates to indent with tabs.",
    )
    cliParser.add_argument(
        "--semicolons",
        dest="semicolons",
        action="store_true",
        default=False,
        help="Whether to end nodes with semicolons or not.",
    )
    cliParser.add_argument(
        "--radix",
        dest="respectRadix",
        action="store_true",
        help="Output numeric values in the radix used by the input. (0x1a outputs as 0x1a)",
    )
    cliParser.add_argument(
        "--no-radix",
        dest="respectRadix",
        action="store_false",
        help="Convert all numeric arguments to decimal. (0x1a outputs as 26)",
    )
    cliParser.add_argument(
        "--raw-strings",
        dest="respectStringType",
        action="store_true",
        help="Output string values in the string type used by the input.",
    )
    cliParser.add_argument(
        "--no-raw-strings",
        dest="respectStringType",
        action="store_false",
        help="Convert all string arguments into plain strings.",
    )
    cliParser.add_argument(
        "--exponent",
        dest="exponent",
        type=expFromString,
        action="store",
        default="e",
        help="What character to use ('e' or 'E') for indicating exponents on scinot numbers.",
    )
    cliParser.set_defaults(respectRadix=True, respectStringType=True)
    options = cliParser.parse_args()
    parseConfig = parsing.ParseConfig(
        nativeUntaggedValues=False,
        nativeTaggedValues=False,
    )
    printConfig = printing.PrintConfig(
        indent=" " * options.indent if options.indent >= 0 else "\t",
        semicolons=options.semicolons,
        respectRadix=options.respectRadix,
        respectStringType=options.respectStringType,
        exponent=options.exponent,
    )

    with options.infile as fh:
        doc = parsefuncs.parse(fh.read(), parseConfig)
    with options.outfile as fh:
        fh.write(doc.print(printConfig))


def expFromString(s: str) -> str:
    if s in ("e", "E"):
        return s
    msg = f"Expected 'e' or 'E' for an exponent; got '{s}'"
    raise argparse.ArgumentTypeError(msg)



from __future__ import annotations

from dataclasses import dataclass

from . import t


@dataclass
class Stream:
    _chars: str
    config: t.ParseConfig

    def __getitem__(self, key: int | slice) -> str:
        try:
            return self._chars[key]
        except IndexError:
            return ""

    def eof(self, index: int) -> bool:
        return index >= len(self._chars)



from __future__ import annotations

import dataclasses
from abc import ABCMeta, abstractmethod
from collections import OrderedDict
from dataclasses import dataclass

from . import printing, t

if t.TYPE_CHECKING:
    VT = t.TypeVar("VT")


@dataclass
class Document:
    nodes: list[Node] = dataclasses.field(default_factory=list)
    printConfig: t.PrintConfig | None = None

    def print(self, config: t.PrintConfig | None = None) -> str:
        config = config or self.printConfig or printing.defaults
        s = ""
        for node in self.nodes:
            node = toKdlNode(node)
            s += node.print(config, 0)
        if s == "":
            # always end a kdl doc with a newline
            s = "\n"
        return s

    @t.overload
    def get(self, key: t.NodeKey) -> Node | None:
        ...

    @t.overload
    def get(self, key: t.NodeKey, default: VT) -> Node | VT:
        ...

    def get(
        self,
        key: t.NodeKey,
        default: VT | None = None,
    ) -> Node | VT | None:
        try:
            return self[key]
        except KeyError:
            return default

    def __getitem__(self, key: t.NodeKey) -> Node:
        tag, name, passedTag = tupleFromNodeKey(key)
        for node in self.nodes:
            if name is not None and node.name != name:
                continue
            if passedTag and node.tag != tag:
                continue
            return node
        if not passedTag:
            msg = name
        elif name is None:
            msg = f"({tag})"
        else:
            msg = f"({tag}){name}"
        raise KeyError(msg)

    def getAll(
        self,
        key: t.NodeKey,
    ) -> t.Iterable[Node]:
        tag, name, passedTag = tupleFromNodeKey(key)
        for node in self.nodes:
            if name is not None and node.name != name:
                continue
            if passedTag and node.tag != tag:
                continue
            yield node

    def __str__(self) -> str:
        return self.print()


@dataclass
class Node:
    name: str
    tag: str | None = None
    args: list[t.Any] = dataclasses.field(default_factory=list)
    props: dict[str, t.Any] = dataclasses.field(default_factory=OrderedDict)
    nodes: list[Node] = dataclasses.field(default_factory=list)

    def print(
        self,
        config: t.PrintConfig | None = None,
        indentLevel: int = 0,
    ) -> str:
        if config is None:
            config = printing.defaults

        s = config.indent * indentLevel

        if self.tag is not None:
            s += f"({printIdent(self.tag)})"

        s += printIdent(self.name)

        # Print all the args, then all the properties
        # in alpha order, using only the last if a key
        # is duplicated.
        for arg in self.args:
            arg = toKdlValue(arg)
            if not config.printNullArgs and isinstance(arg, Null):
                continue
            s += f" {arg.print(config)}"

        props: t.Iterable[tuple[str, t.Any]]
        if config.sortProperties:
            props = sorted(self.props.items(), key=lambda x: x[0])
        else:
            props = self.props.items()
        for key, value in props:
            value = toKdlValue(value)
            if not config.printNullProps and isinstance(value, Null):
                continue
            s += f" {printIdent(key)}={value.print(config)}"

        if self.nodes:
            childrenText = ""
            for child in self.nodes:
                child = toKdlNode(child)
                childrenText += child.print(config=config, indentLevel=indentLevel + 1)
            if childrenText:
                s += " {\n"
                s += childrenText
                s += config.indent * indentLevel + "}"
        if config.semicolons:
            s += ";\n"
        else:
            s += "\n"
        return s

    @t.overload
    def get(self, key: t.NodeKey) -> Node | None:
        ...

    @t.overload
    def get(self, key: t.NodeKey, default: VT) -> Node | VT:
        ...

    def get(
        self,
        key: t.NodeKey,
        default: VT | None = None,
    ) -> Node | VT | None:
        try:
            return self[key]
        except KeyError:
            return default

    def __getitem__(self, key: t.NodeKey) -> Node:
        tag, name, passedTag = tupleFromNodeKey(key)
        for node in self.nodes:
            if name is not None and node.name != name:
                continue
            if passedTag and node.tag != tag:
                continue
            return node
        if not passedTag:
            msg = name
        elif name is None:
            msg = f"({tag})"
        else:
            msg = f"({tag}){name}"
        raise KeyError(msg)

    def getAll(
        self,
        key: t.NodeKey,
    ) -> t.Iterable[Node]:
        tag, name, passedTag = tupleFromNodeKey(key)
        for node in self.nodes:
            if name is not None and node.name != name:
                continue
            if passedTag and node.tag != tag:
                continue
            yield node

    def __str__(self) -> str:
        return self.print()


class Value(metaclass=ABCMeta):
    value: t.Any
    tag: str | None

    @abstractmethod
    def print(self, config: t.PrintConfig | None = None) -> str:
        pass


@dataclass
class ExactValue(Value):
    # Not produced by anything in the parser,
    # but used when a native type needs a precise output
    # not captured by the native semantics,
    # like precise number formatting.
    # .chars *must* be a valid KDL value
    chars: str
    tag: str | None = None

    def print(self, config: t.PrintConfig | None = None) -> str:
        return printTag(self.tag) + self.chars

    def __str__(self) -> str:
        return self.print()


class Numberish(Value, metaclass=ABCMeta):
    pass


@dataclass
class Binary(Numberish):
    value: int
    tag: str | None = None

    def print(self, config: t.PrintConfig | None = None) -> str:
        if config is None:
            config = printing.defaults
        s = printTag(self.tag)
        if config.respectRadix:
            s += bin(self.value)
        else:
            s += str(self.value)
        return s

    def __str__(self) -> str:
        return self.print()


@dataclass
class Octal(Numberish):
    value: int
    tag: str | None = None

    def print(self, config: t.PrintConfig | None = None) -> str:
        if config is None:
            config = printing.defaults
        s = printTag(self.tag)
        if config.respectRadix:
            s += oct(self.value)
        else:
            s += str(self.value)
        return s

    def __str__(self) -> str:
        return self.print()


@dataclass
class Decimal(Numberish):
    mantissa: int | float
    exponent: int = 0
    tag: str | None = None

    @property
    def value(self) -> float:
        return self.mantissa * (10.0**self.exponent)

    def print(self, config: t.PrintConfig | None = None) -> str:
        if config is None:
            config = printing.defaults
        s = printTag(self.tag) + str(self.mantissa)
        if self.exponent != 0:
            s += config.exponent
            if self.exponent > 0:
                s += "+"
            s += str(self.exponent)
        return s

    def __str__(self) -> str:
        return self.print()


@dataclass
class Hex(Numberish):
    value: int
    tag: str | None = None

    def print(self, config: t.PrintConfig | None = None) -> str:
        if config is None:
            config = printing.defaults
        s = printTag(self.tag)
        if config.respectRadix:
            s += hex(self.value)
        else:
            s += str(self.value)
        return s

    def __str__(self) -> str:
        return self.print()


@dataclass
class Bool(Value):
    value: bool
    tag: str | None = None

    def print(self, config: t.PrintConfig | None = None) -> str:
        if config is None:
            config = printing.defaults
        return printTag(self.tag) + ("true" if self.value else "false")

    def __str__(self) -> str:
        return self.print()


@dataclass
class Null(Value):
    tag: str | None = None

    @property
    def value(self) -> None:
        return None

    def print(self, config: t.PrintConfig | None = None) -> str:
        if config is None:
            config = printing.defaults
        return printTag(self.tag) + "null"

    def __str__(self) -> str:
        return self.print()


class Stringish(Value, metaclass=ABCMeta):
    pass


@dataclass
class RawString(Stringish):
    value: str
    tag: str | None = None

    def print(self, config: t.PrintConfig | None = None) -> str:
        if config is None:
            config = printing.defaults
        if config.respectStringType:
            hashes = "#" * findRequiredHashCount(self.value)
            return f'{printTag(self.tag)}r{hashes}"{self.value}"{hashes}'
        else:
            return f'{printTag(self.tag)}"{escapedFromRaw(self.value)}"'

    def __str__(self) -> str:
        return self.print()


def findRequiredHashCount(chars: str) -> int:
    for i in range(0, 100):
        ender = '"' + ("#" * i)
        if ender not in chars:
            return i
    assert False, "A raw string requires more than 100 hashes???"


@dataclass
class String(Stringish):
    value: str
    tag: str | None = None

    def print(self, config: t.PrintConfig | None = None) -> str:
        if config is None:
            config = printing.defaults
        return f'{printTag(self.tag)}"{escapedFromRaw(self.value)}"'

    def __str__(self) -> str:
        return self.print()


def toKdlNode(val: t.Any) -> Node:
    if isinstance(val, Node):
        return val
    if not callable(getattr(val, "to_kdl", None)):
        msg = f"Can't convert object to KDL for serialization. Got:\n{val!r}"
        raise Exception(
            msg,
        )
    node = val.to_kdl()
    if not isinstance(node, Node):
        msg = f"Expected object to convert to KDL Node. Got:\n{val!r}"
        raise Exception(msg)
    return node


def toKdlValue(val: t.Any) -> t.KDLValue:
    """
    Converts any KDLish value (a KDLValue, a primitive,
    an object corresponding to a built-in tag,
    or an object with .to_kdl() that returns one of the above)
    into a KDLValue
    """
    import base64
    import datetime
    import decimal
    import ipaddress
    import re
    import urllib
    import uuid

    if isinstance(val, Value):
        return val
    if val is None:
        return Null()
    if isinstance(val, bool):
        return Bool(val)
    if isinstance(val, str):
        return String(val)
    if isinstance(val, (int, float)):
        return Decimal(val)
    if isinstance(val, decimal.Decimal):
        return String(str(val), "decimal")
    if isinstance(val, datetime.datetime):
        return String(val.isoformat(), "date-time")
    if isinstance(val, datetime.time):
        return String(val.isoformat(), "time")
    if isinstance(val, datetime.date):
        return String(val.isoformat(), "date")
    if isinstance(val, ipaddress.IPv4Address):
        return String(str(val), "ipv4")
    if isinstance(val, ipaddress.IPv6Address):
        return String(str(val), "ipv6")
    if isinstance(val, urllib.parse.ParseResult):
        return String(urllib.parse.urlunparse(val), "url")
    if isinstance(val, uuid.UUID):
        return String(str(val), "uuid")
    if isinstance(val, re.Pattern):
        return RawString(val.pattern, "regex")
    if isinstance(val, bytes):
        return String(base64.b64encode(val).decode("utf-8"), "base-64")

    if not callable(getattr(val, "to_kdl", None)):
        msg = f"Can't convert object to KDL for serialization. Got:\n{val!r}"
        raise Exception(
            msg,
        )

    convertedVal = val.to_kdl()
    if isKdlishValue(convertedVal):
        return toKdlValue(convertedVal)
    else:
        msg = f"Expected object to convert to KDL value or compatible primitive. Got:\n{val!r}"
        raise Exception(msg)


def isKdlishValue(val: t.Any) -> bool:
    import datetime
    import decimal
    import ipaddress
    import re
    import urllib
    import uuid

    if val is None:
        return True
    return isinstance(
        val,
        (
            str,
            int,
            float,
            bool,
            decimal.Decimal,
            datetime.time,
            datetime.date,
            datetime.datetime,
            ipaddress.IPv4Address,
            ipaddress.IPv6Address,
            urllib.parse.ParseResult,
            uuid.UUID,
            re.Pattern,
            bytes,
            Value,
        ),
    )


def printTag(tag: str | None) -> str:
    if tag is not None:
        return f"({printIdent(tag)})"
    else:
        return ""


def tupleFromNodeKey(key: t.NodeKey) -> tuple[str | None, str | None, bool]:
    if isinstance(key, str):
        return None, key, False
    return key[0], key[1], True


def escapedFromRaw(chars: str) -> str:
    return (
        chars.replace("\\", "\\\\")
        .replace('"', '\\"')
        # don't escape a forward slash when printing
        .replace("\b", "\\b")
        .replace("\f", "\\f")
        .replace("\n", "\\n")
        .replace("\r", "\\r")
        .replace("\t", "\\t")
    )


def printIdent(chars: str) -> str:
    if isBareIdent(chars):
        return chars
    return f'"{escapedFromRaw(chars)}"'


def isBareIdent(chars: str) -> bool:
    if not chars:
        return False
    if any(not isIdentChar(x) for x in chars):
        return False
    if chars[0] in "0123456789":
        return False
    if len(chars) > 1 and chars[0] in "+-" and chars[1] in "0123456789":
        return False
    if chars in ("true", "false", "null"):
        return False
    return True


def isIdentChar(ch: str) -> bool:
    if not ch:
        return False
    # reserved characters
    if ch in r'''\/(){}<>;[]=,"''':
        return False
    cp = ord(ch)
    # nonprintable
    if cp < 0x20:
        return False
    # invalid codepoint
    if cp > 0x10FFFF:
        return False
    # spaces
    if cp in (0x09, 0x20, 0xA0, 0x1680, 0x202F, 0x205F, 0x3000, 0xFEFF):
        return False
    # more spaces
    if 0x2000 <= cp <= 0x200A:
        return False
    # line breaks
    if cp in (0x0A, 0x0D, 0x85, 0x0C, 0x2028, 0x2029):
        return False
    return True



from __future__ import annotations

from dataclasses import dataclass, field

from . import parsefuncs, t


class Parser:
    def __init__(
        self,
        parseConfig: ParseConfig | None = None,
        printConfig: t.PrintConfig | None = None,
    ) -> None:
        self.parseConfig = parseConfig
        self.printConfig = printConfig

    def parse(self, chars: str, config: ParseConfig | None = None) -> t.Document:
        doc = parsefuncs.parse(chars, config or self.parseConfig or defaults)
        doc.printConfig = self.printConfig
        return doc


@dataclass
class ParseConfig:
    nativeUntaggedValues: bool = True
    nativeTaggedValues: bool = True
    valueConverters: dict[str, t.Callable] = field(default_factory=dict)
    nodeConverters: dict[str | tuple[str, str], t.Callable] = field(
        default_factory=dict,
    )


defaults = ParseConfig()



from __future__ import annotations

from . import converters, parsing, t, types
from .errors import ParseError, ParseFragment
from .result import Failure, Result
from .stream import Stream


def parse(input: str, config: t.ParseConfig | None = None) -> t.Document:
    if config is None:
        config = parsing.defaults
    doc = types.Document()
    s = Stream(input, config)
    _, i = parseLinespace(s, 0)
    while not s.eof(i):
        node, i = parseNode(s, i)
        if node is Failure:
            raise ParseError(s, i, "Expected a node")
        if node:
            doc.nodes.append(node)
        _, i = parseLinespace(s, i)
    return doc


def parseNode(s: Stream, start: int) -> Result:
    i = start

    sd, i = parseSlashDash(s, i)
    if sd is Failure:
        sd = False

    # tag?
    tag, i = parseTag(s, i)
    if tag is Failure:
        tag = None

    # name
    name, i = parseIdent(s, i)
    if name is Failure:
        return Result.fail(start)

    nameEnd = i

    node = types.Node(tag=tag, name=name)

    # props and args
    while True:
        space, i = parseNodespace(s, i)
        if space is Failure:
            break
        entity, i = parseEntity(s, i)
        if entity is Failure:
            break
        if entity is None:
            continue
        if entity[0] is None:
            node.args.append(entity[1])
        else:
            node.props[entity[0]] = entity[1]

    _, i = parseNodespace(s, i)

    nodes, i = parseNodeChildren(s, i)
    if nodes is not Failure:
        node.nodes = nodes

    _, i = parseNodespace(s, i)
    _, i = parseNodeTerminator(s, i)

    if sd:
        return Result(None, i)
    else:
        if (tag, name) in s.config.nodeConverters:
            node = s.config.nodeConverters[tag, name](
                node,
                ParseFragment(s[start:nameEnd], s, start),
            )
        elif name in s.config.nodeConverters:
            node = s.config.nodeConverters[name](
                node,
                ParseFragment(s[start:nameEnd], s, start),
            )
        return Result(node, i)


def parseNodeChildren(s: Stream, start: int) -> Result:
    sd, i = parseSlashDash(s, start)
    if sd is Failure:
        sd = False

    if s[i] != "{":
        return Result.fail(start)
    i += 1
    nodes = []
    while True:
        _, i = parseLinespace(s, i)
        node, i = parseNode(s, i)
        if node is Failure:
            break
        if node is not None:
            nodes.append(node)
    _, i = parseLinespace(s, i)
    if s.eof(i):
        raise ParseError(s, start, "Hit EOF while searching for end of child list")
    if s[i] != "}":
        raise ParseError(s, i, "Junk between end of child list and closing }")
    if sd:
        return Result(None, i + 1)
    else:
        return Result(nodes, i + 1)


def parseTag(s: Stream, start: int) -> Result:
    if s[start] != "(":
        return Result.fail(start)
    tag, end = parseIdent(s, start + 1)
    if tag is Failure:
        return Result.fail(start)
    if s[end] != ")":
        raise ParseError(s, end, "Junk between tag ident and closing paren.")
    return Result(tag, end + 1)


def parseIdent(s: Stream, start: int) -> Result:
    string, i = parseString(s, start)
    if string is not Failure:
        return Result(string.value, i)
    return parseBareIdent(s, start)


def parseBareIdent(s: Stream, start: int) -> Result:
    res, i = parseIdentStart(s, start)
    if res is Failure:
        return Result.fail(start)
    while isIdentChar(s[i]):
        i += 1
    ident = s[start:i]
    if isKeyword(ident):
        return Result.fail(start)
    return Result(ident, i)


def parseIdentStart(s: Stream, start: int) -> Result:
    if isDigit(s[start]) or (isSign(s[start]) and isDigit(s[start + 1])):
        return Result.fail(start)
    if not isIdentChar(s[start]):
        return Result.fail(start)
    return Result(s[start], start + 1)


def parseNodeTerminator(s: Stream, start: int) -> Result:
    res = parseNewline(s, start)
    if res.valid:
        return res
    res = parseSingleLineComment(s, start)
    if res.valid:
        return res
    if s[start] == ";":
        return Result(";", start + 1)
    if s.eof(start):
        return Result(True, start)
    raise ParseError(s, start, "Junk after node, before terminator.")


def parseEntity(s: Stream, start: int) -> Result:
    sd, i = parseSlashDash(s, start)
    if sd is Failure:
        sd = False

    ent, i = parseProperty(s, i)
    if ent is Failure:
        ent, i = parseValue(s, i)
        if ent is Failure:
            return Result.fail(start)
    if sd:
        return Result(None, i)
    else:
        return Result(ent, i)


def parseProperty(s: Stream, start: int) -> Result:
    key, i = parseIdent(s, start)
    if key is Failure:
        return Result.fail(start)
    if s[i] != "=":
        # property name might be a string,
        # so this isn't point-of-no-return yet
        return Result.fail(start)
    entity, i = parseValue(s, i + 1)
    if entity is Failure:
        raise ParseError(s, i, "Expected value after prop=.")
    return Result((key, entity[1]), i)


def parseValue(s: Stream, start: int) -> Result:
    tag, i = parseTag(s, start)
    if tag is Failure:
        tag = None

    valueStart = i
    val, i = parseNumber(s, i)
    if val is Failure:
        val, i = parseKeyword(s, i)
        if val is Failure:
            val, i = parseString(s, i)
    if val is not Failure:
        if tag is None and s.config.nativeUntaggedValues:
            val = val.value
        else:
            val.tag = tag
        if tag is not None and tag in s.config.valueConverters:
            val = s.config.valueConverters[tag](
                val,
                ParseFragment(s[valueStart:i], s, i),
            )
        if tag is not None and s.config.nativeTaggedValues and isinstance(val, types.Value):
            val = converters.toNative(val, ParseFragment(s[valueStart:i], s, i))
        return Result((None, val), i)

    if s[i] == "'":
        raise ParseError(s, i, "KDL strings use double-quotes.")

    ident, _ = parseBareIdent(s, i)
    if ident is not Failure and ident.lower() in ("true", "false", "null"):
        raise ParseError(s, i, "KDL keywords are lower-case.")

    # Failed to find a value
    # But if I found a tag, something's up
    if tag is not None:
        raise ParseError(s, i, "Found a tag, but no value following it.")
    return Result.fail(start)


def parseNumber(s: Stream, start: int) -> Result:
    if not isNumberStart(s, start):
        return Result.fail(start)
    res = parseBinaryNumber(s, start)
    if res.valid:
        return res
    res = parseOctalNumber(s, start)
    if res.valid:
        return res
    res = parseHexNumber(s, start)
    if res.valid:
        return res
    res = parseDecimalNumber(s, start)
    if res.valid:
        return res
    raise ParseError(
        s,
        start,
        "Expected a number, but got junk after the initial digit.",
    )


def parseBinaryNumber(s: Stream, start: int) -> Result:
    i = start

    # optional sign
    sign, i = parseSign(s, i)
    if sign is Failure:
        sign = 1

    # prefix
    if not (s[i] == "0" and s[i + 1] == "b"):
        return Result.fail(start)
    i += 2

    # initial digit
    if not isBinaryDigit(s[i]):
        raise ParseError(s, i, "Expected binary digit after 0b, got junk.")

    # following digits/underscores
    end = i + 1
    while isBinaryDigit(s[end]) or s[end] == "_":
        end += 1
    value = int(s[i:end].replace("_", ""), 2) * sign
    return Result(types.Binary(value), end)


def parseOctalNumber(s: Stream, start: int) -> Result:
    i = start

    # optional sign
    sign, i = parseSign(s, i)
    if sign is Failure:
        sign = 1

    # prefix
    if not (s[i] == "0" and s[i + 1] == "o"):
        return Result.fail(start)
    i += 2

    # initial digit
    if not isOctalDigit(s[i]):
        raise ParseError(s, i, "Expected octal digit after 0o, got junk.")

    # following digits/underscores
    end = i + 1
    while isOctalDigit(s[end]) or s[end] == "_":
        end += 1
    value = int(s[i:end].replace("_", ""), 8) * sign
    return Result(types.Octal(value), end)


def parseHexNumber(s: Stream, start: int) -> Result:
    i = start

    # optional sign
    sign, i = parseSign(s, i)
    if sign is Failure:
        sign = 1

    # prefix
    if not (s[i] == "0" and s[i + 1] == "x"):
        return Result.fail(start)
    i += 2

    # initial digit
    if not isHexDigit(s[i]):
        raise ParseError(s, i, "Expected hex digit after 0x, got junk.")

    # following digits/underscores
    end = i + 1
    while isHexDigit(s[end]) or s[end] == "_":
        end += 1
    value = int(s[i:end].replace("_", ""), 16) * sign
    return Result(types.Hex(value), end)


def parseDecimalNumber(s: Stream, start: int) -> Result:
    i = start

    # optional sign
    _, i = parseSign(s, i)

    # integer part
    _, i = parseDigits(s, i)

    if s[i] == ".":
        result, i = parseDigits(s, i + 1)
        if result is Failure:
            raise ParseError(s, i, "Expected digit after decimal point.")

    mantissaChars = s[start:i].replace("_", "")
    mantissa: int | float
    try:
        mantissa = int(mantissaChars, 10)
    except ValueError:
        try:
            mantissa = float(mantissaChars)
        except ValueError as exc:
            raise ParseError(
                s,
                start,
                "Number-like string didn't actually parse as a number.",
            ) from exc

    exponent = 0
    if s[i] in ("e", "E"):
        exponentStart = i + 1
        _, i = parseSign(s, i + 1)
        result, i = parseDigits(s, i)
        if result is Failure:
            raise ParseError(s, i, "Expected number after exponent.")
        exponent = int(s[exponentStart:i].replace("_", ""))

    return Result(types.Decimal(mantissa, exponent), i)


def parseDigits(s: Stream, start: int) -> Result:
    # First digit must be decimal digit
    # Subsequent can be digit or underscore
    if not isDigit(s[start]):
        return Result.fail(start)
    end = start + 1
    while isDigit(s[end]) or s[end] == "_":
        end += 1
    return Result(True, end)


def isNumberStart(s: Stream, start: int) -> bool:
    # all numbers start with an optional sign
    # followed by a digit:
    # either the first digit of the number,
    # or the 0 of the prefix)
    if isDigit(s[start]):
        return True
    if isSign(s[start]) and isDigit(s[start + 1]):
        return True
    return False


def parseSign(s: Stream, start: int) -> Result:
    if s[start] == "+":
        return Result(1, start + 1)
    if s[start] == "-":
        return Result(-1, start + 1)
    return Result.fail(start)


def parseKeyword(s: Stream, start: int) -> Result:
    if s[start : start + 4] == "true" and not isIdentChar(s[start + 4]):
        return Result(types.Bool(True), start + 4)
    if s[start : start + 5] == "false" and not isIdentChar(s[start + 5]):
        return Result(types.Bool(False), start + 5)
    if s[start : start + 4] == "null" and not isIdentChar(s[start + 4]):
        return Result(types.Null(), start + 4)
    return Result.fail(start)


def parseString(s: Stream, start: int) -> Result:
    if s[start] == '"':
        return parseEscapedString(s, start)
    if s[start] == "r" and s[start + 1] in ("#", '"'):
        return parseRawString(s, start)
    return Result.fail(start)


def parseEscapedString(s: Stream, start: int) -> Result:
    if s[start] != '"':
        return Result.fail(start)
    i = start + 1

    rawChars = ""
    while True:
        if s[i] not in ("\\", '"', ""):
            rawChars += s[i]
            i += 1
            continue
        if s[i] == '"':
            break
        if s[i] == "":
            raise ParseError(
                s,
                start,
                "Hit EOF while looking for the end of the string",
            )
        ch, i = parseEscape(s, i)
        if ch is Failure:
            raise ParseError(s, i, "Invalid escape sequence in string")
        rawChars += ch

    return Result(types.String(rawChars), i + 1)


def parseEscape(s: Stream, start: int) -> Result:
    if s[start] != "\\":
        return Result.fail(start)
    ch = s[start + 1]
    if ch == "n":
        return Result("\n", start + 2)
    if ch == "r":
        return Result("\r", start + 2)
    if ch == "t":
        return Result("\t", start + 2)
    if ch == "\\":
        return Result("\\", start + 2)
    if ch == "/":
        return Result("/", start + 2)
    if ch == '"':
        return Result('"', start + 2)
    if ch == "b":
        return Result("\b", start + 2)
    if ch == "f":
        return Result("\f", start + 2)
    if ch == "u":
        if s[start + 2] != "{":
            raise ParseError(
                s,
                start,
                "Unicode escapes must surround their codepoint in {}",
            )
        i = start + 3
        hexStart = i
        while isHexDigit(s[i]):
            i += 1
        hexCount = i - hexStart
        if s[i] != "}":
            raise ParseError(s, hexStart, "Expected } to finish a unicode escape")
        if hexCount < 1:
            raise ParseError(s, hexStart, "Unicode escape doesn't contain a codepoint")
        if hexCount > 6:
            raise ParseError(
                s,
                hexStart,
                "Unicode escapes can contain at most six digits",
            )
        hexValue = int(s[hexStart:i], 16)
        if hexValue > 0x10FFFF:
            raise ParseError(
                s,
                hexStart,
                "Maximum codepoint in a unicode escape is 0x10ffff",
            )
        return Result(chr(hexValue), i + 1)
    raise ParseError(s, start, "Invalid character escape")


def parseRawString(s: Stream, start: int) -> Result:
    if s[start] != "r":
        return Result.fail(start)
    i = start + 1

    # count hashes
    hashCount, i = parseInitialHashes(s, i)

    if s[i] != '"':
        return Result.fail(start)
    i = i + 1

    stringStart = i
    while True:
        while s[i] not in ('"', ""):
            i += 1
        if s[i] == "":
            raise ParseError(
                s,
                start,
                "Hit EOF while looking for the end of the raw string.",
            )
        stringEnd = i
        i += 1
        result, i = parseFinalHashes(s, i, expectedCount=hashCount)
        if result is not Failure:
            return Result(types.RawString(s[stringStart:stringEnd]), i)


def parseInitialHashes(s: Stream, start: int) -> Result:
    i = start
    while s[i] == "#":
        i += 1
    if s[i] == '"':
        return Result(i - start, i)
    return Result.fail(start)


def parseFinalHashes(s: Stream, start: int, expectedCount: int) -> Result:
    i = start
    while s[i] == "#":
        i += 1
    count = i - start
    if count < expectedCount:
        return Result.fail(start)
    if count > expectedCount:
        raise ParseError(
            s,
            start,
            f"Expected {expectedCount} hashes at end of raw string; got {count}.",
        )
    return Result(True, i)


def parseNewline(s: Stream, start: int) -> Result:
    if s[start] == "\x0d" and s[start + 1] == "\x0a":
        return Result(True, start + 2)
    if isNewlineChar(s[start]):
        return Result(True, start + 1)
    return Result.fail(start)


def parseLinespace(s: Stream, start: int) -> Result:
    i = start
    while True:
        nl, i = parseNewline(s, i)
        ws, i = parseWhitespace(s, i)
        sc, i = parseSingleLineComment(s, i)
        if nl is Failure and ws is Failure and sc is Failure:
            break
    if i == start:
        return Result.fail(start)
    return Result(True, i)


def parseNodespace(s: Stream, start: int) -> Result:
    i = start
    while True:
        _, i = parseWhitespace(s, i)
        escline, i = parseEscline(s, i)
        if escline is Failure:
            break
    if i == start:
        return Result.fail(start)
    return Result(True, i)


def parseEscline(s: Stream, start: int) -> Result:
    if s[start] != "\\":
        return Result.fail(start)
    _, i = parseWhitespace(s, start + 1)
    if s[i] == "\n":
        return Result(True, i + 1)
    sl, i = parseSingleLineComment(s, i)
    if sl is not Failure:
        return Result(True, i)
    return Result.fail(start)


def parseWhitespace(s: Stream, start: int) -> Result:
    i = start
    while True:
        sp, i = parseUnicodeSpace(s, i)
        bc, i = parseBlockComment(s, i)
        if sp is Failure and bc is Failure:
            break
    if i == start:
        return Result.fail(start)
    return Result(True, i)


def parseUnicodeSpace(s: Stream, start: int) -> Result:
    if not isWSChar(s[start]):
        return Result.fail(start)
    end = start + 1
    while isWSChar(s[end]):
        end += 1
    return Result(True, end)


def isIdentChar(ch: str) -> bool:
    if not ch:
        return False
    if ch in r'''\/(){}<>;[]=,"''':
        return False
    if isLinespaceChar(ch):
        return False
    cp = ord(ch)
    if cp <= 0x20:
        return False
    if cp > 0x10FFFF:
        return False
    return True


def isKeyword(ident: str) -> bool:
    return ident in ("null", "true", "false")


def isSign(ch: str) -> bool:
    return ch != "" and ch in "+-"


def isDigit(ch: str) -> bool:
    return ch != "" and ch in "0123456789"


def isBinaryDigit(ch: str) -> bool:
    return ch != "" and ch in "01"


def isOctalDigit(ch: str) -> bool:
    return ch != "" and ch in "01234567"


def isHexDigit(ch: str) -> bool:
    return ch != "" and ch in "0123456789abcdefABCDEF"


def isWSChar(ch: str) -> bool:
    if not ch:
        return False
    cp = ord(ch)
    if cp in (0x09, 0x20, 0xA0, 0x1680, 0x202F, 0x205F, 0x3000, 0xFEFF):
        return True
    if 0x2000 <= cp <= 0x200A:
        return True
    return False


def isNewlineChar(ch: str) -> bool:
    if not ch:
        return False
    cp = ord(ch)
    return cp in (0x0A, 0x0D, 0x85, 0x0C, 0x2028, 0x2029)


def isLinespaceChar(ch: str) -> bool:
    if not ch:
        return False
    return isWSChar(ch) or isNewlineChar(ch)


def parseSlashDash(s: Stream, start: int) -> Result:
    if s[start] == "/" and s[start + 1] == "-":
        _, i = parseNodespace(s, start + 2)
        return Result(True, i)
    return Result.fail(start)


def parseSingleLineComment(s: Stream, start: int) -> Result:
    if not (s[start] == "/" and s[start + 1] == "/"):
        return Result.fail(start)
    i = start + 2
    while not isNewlineChar(s[i]) and not s.eof(i):
        i += 1
    _, i = parseNewline(s, i)
    return Result(True, i)


def parseBlockComment(s: Stream, start: int) -> Result:
    if not (s[start] == "/" and s[start + 1] == "*"):
        return Result.fail(start)
    i = start + 2
    while True:
        if s.eof(i):
            raise ParseError(s, start, "Hit EOF while inside a multiline comment")
        if s[i] == "*" and s[i + 1] == "/":
            return Result(True, i + 2)
        if s[i] == "/" and s[i + 1] == "*":
            _, i = parseBlockComment(s, i)
            continue
        i += 1



from dataclasses import dataclass


@dataclass
class PrintConfig:
    indent: str = "\t"
    semicolons: bool = False
    printNullArgs: bool = True
    printNullProps: bool = True
    respectRadix: bool = True
    respectStringType: bool = True
    exponent: str = "e"
    sortProperties: bool = False


defaults = PrintConfig()



from __future__ import annotations

from . import t


class ParseError(Exception):
    def __init__(self, s: str | t.Stream, i: int, msg: str) -> None:
        self.line, self.col = lineCol(s, i)
        self.msg = f"{self.line}:{self.col} parse error:"
        if "\n" in msg:
            self.msg += "\n" + msg
        elif len(self.msg) + len(msg) + 1 > 78:
            self.msg += "\n  " + msg
        else:
            self.msg += " " + msg
        super().__init__(self.msg)


def lineCol(s: str | t.Stream, index: int) -> tuple[int, int]:
    """Determines the line and column from an index."""
    line = 1
    col = 1
    for i in range(index):
        if s[i] == "\n":
            line += 1
            col = 1
            continue
        col += 1
    return line, col


class ParseFragment:
    def __init__(self, fragment: str, s: t.Stream, i: int) -> None:
        self.fragment = fragment
        self._s = s
        self._i = i

    def error(self, msg: str) -> ParseError:
        return ParseError(self._s, self._i, msg)
